"""Kasa Cloud API module for PyKasaCloud."""

from collections.abc import Callable, Coroutine
from dataclasses import dataclass
import json
from typing import TYPE_CHECKING, Any, TypedDict, cast
import uuid

from aiohttp import ClientSession
import anyio
from kasa import (
    AuthenticationError,
    BaseTransport,
    DeviceConfig,
    DeviceError,
    KasaException,
)
from kasa.exceptions import _RetryableError
from kasa.httpclient import HttpClient
from kasa.json import dumps as json_dumps, loads as json_loads
from yarl import URL

from pykasacloud.const import (
    API_URL,
    APPSERVERURL,
    APPTYPE,
    CLIENT_ID,
    REFRESH_TOKEN,
    TOKEN,
    USERAGENT,
)
from pykasacloud.exceptions import CloudErrorCode, KasaCloudError

# _LOGGER = logging.getLogger(__name__)


@dataclass
class Token(TypedDict):
    """Token dataclass."""

    token: str
    refresh_token: str
    client_id: str
    account_id: int


class CloudTransport(BaseTransport):
    """CloudTransport."""

    COMMON_HEADERS = {"User-Agent": USERAGENT}

    _token: Token
    _token_storage_file: str | None
    _token_update_callback: Callable[[Token], Coroutine] | None
    _url: URL = URL(API_URL)
    _http_client: HttpClient

    @classmethod
    async def auth(
        cls,
        *,
        client_session: ClientSession,
        username: str | None = None,
        password: str | None = None,
        token: Token | None = None,
        token_storage_file: str | None = None,
        token_update_callback: Callable[[Token], Coroutine] | None = None,
    ) -> "CloudTransport":
        """Create a CloudTransport Class."""

        self = cls(
            config=DeviceConfig(host="TPLink/Kasa Cloud", http_client=client_session),
        )

        self._token_storage_file = token_storage_file
        self._token_update_callback = token_update_callback

        self._http_client = HttpClient(config=self._config)

        if not token:
            if not username:
                if not token_storage_file:
                    raise AuthenticationError(
                        "Username/password or tokens must be provided"
                    )
                # Try and load from file cache
                try:
                    async with await anyio.open_file(
                        token_storage_file, "r", encoding="utf-8"
                    ) as f:
                        self._token = Token(**json.loads(await f.read()))
                        await f.aclose()

                except FileNotFoundError as ex:
                    raise AuthenticationError(
                        "Username/password or tokens must be provided"
                    ) from ex

            else:
                if not password:
                    raise AuthenticationError(
                        "Username provided but password is missing"
                    )
                # login for the first time
                client_id: str = str(uuid.uuid4())
                payload: dict[str, Any] = {
                    "method": "login",
                    "params": {
                        "cloudUserName": username,
                        "cloudPassword": password,
                        "terminalUUID": client_id,
                        "refreshTokenNeeded": "true",
                    },
                }
                auth_results = await self._send_request(payload)
                self._token = Token(
                    token=auth_results["token"],
                    refresh_token=auth_results["refreshToken"],
                    client_id=client_id,
                    account_id=auth_results["accountId"],
                )
                await self._cache_tokens()
        else:
            self._token = token

        return self

    @property
    def token(self) -> Token:
        """Return the current token."""
        return self._token

    async def _cache_tokens(self) -> None:
        if self._token_storage_file:
            async with await anyio.open_file(
                self._token_storage_file, "w", encoding="utf-8"
            ) as f:
                json.dump(self._token, f, ensure_ascii=False, indent=4)
                await f.aclose()
        if self._token_update_callback:
            await self._token_update_callback(self._token)

    async def _handle_cloud_response_error_code(self, resp_dict: Any) -> None:
        """Handle response errors to request reauth etc."""

        try:
            cloud_error_code = CloudErrorCode(resp_dict["error_code"])
        except ValueError as ex:
            raise KasaCloudError(f"Got unknown error: {resp_dict}") from ex

        if cloud_error_code == CloudErrorCode.SUCCESS:
            return

        msg = resp_dict["msg"]
        if cloud_error_code == CloudErrorCode.TOKEN_EXPIRED:
            await self._refresh_token()
            raise _RetryableError(
                "{msg}: {self._host}: {error_code.name}({code})",
                error_code=cloud_error_code,
            )
        if cloud_error_code == CloudErrorCode.DEVICE_OFFLINE:
            raise DeviceError(f"{msg}: {cloud_error_code.value}")
        if cloud_error_code == CloudErrorCode.INVALID_EMAIL_OR_PASSWORD:
            raise AuthenticationError(resp_dict)
        raise KasaCloudError(
            f"{msg}: {cloud_error_code.name}({cloud_error_code.value})"
        )

    async def send(self, request: str) -> dict[str, Any]:
        """Not implemented."""
        return {}

    async def send_request(
        self,
        payload: dict[str, Any],
        device_id: str | None = None,
        url: URL | None = None,
    ) -> dict[str, Any]:
        """Send request to Kasa Cloud."""
        cpayload: dict[str, Any] = {"params": {}}
        cpayload["params"]["token"] = self._token[TOKEN]
        if device_id:
            cpayload["method"] = "passthrough"
            cpayload["params"]["deviceId"] = device_id
            cpayload["params"]["requestData"] = json_dumps(payload)
        else:
            cpayload["method"] = payload["method"]
        return await self._send_request(cpayload, url)

    async def _send_request(
        self, payload: dict[str, Any], url: URL | None = None
    ) -> dict[str, Any]:
        """Send a request."""
        if "params" not in payload:
            payload["params"] = {}
        payload["params"]["appType"] = APPTYPE

        if not url:
            url = self._url

        status_code, resp_dict = await self._http_client.post(
            url=url, json=payload, headers=self.COMMON_HEADERS
        )

        if status_code != 200:
            raise KasaException(
                f"{self._host} responded with an unexpected status code {status_code}"
            )

        # _LOGGER.debug("Response with %s: %r", status_code, resp_dict)

        await self._handle_cloud_response_error_code(resp_dict)

        if TYPE_CHECKING:
            resp_dict = cast(dict[str, Any], resp_dict)

        if "result" in resp_dict:
            resp_dict = resp_dict["result"]
            if resp_dict and "deviceList" in resp_dict and resp_dict["deviceList"]:
                # get the url
                self._url = URL(resp_dict["deviceList"][0][APPSERVERURL])

        if "responseData" in resp_dict:
            resp_dict = resp_dict["responseData"]
            if isinstance(resp_dict, str):
                resp_dict = json_loads(resp_dict)

        # await self.close()

        return resp_dict

    async def _refresh_token(self) -> None:
        if not self._token[REFRESH_TOKEN]:
            raise AuthenticationError(
                "No tokens or refresh token available for refreshing"
            )
        if not self._token[CLIENT_ID]:
            raise AuthenticationError("No client_id available for refreshing token")
        payload: dict[str, Any] = {
            "method": "refreshToken",
            "params": {
                "refreshToken": self._token[REFRESH_TOKEN],
                "terminalUUID": self._token[CLIENT_ID],
            },
        }
        new_token: dict[str, Any] = await self._send_request(payload)
        self._token[TOKEN] = new_token[TOKEN]
        await self._cache_tokens()

    @property
    def default_port(self) -> int:
        """Default port is irrevelant for cloud operations."""
        return 443

    @property
    def credentials_hash(self) -> str | None:
        """Cloud doesn't use credentials hash."""
        return None

    async def close(self) -> None:
        """Close the transport."""
        await self._http_client.close()

    async def reset(self) -> None:
        """This does nothing."""



"""Cloud Protocol."""

import logging
from pprint import pformat as pf
from typing import Any, cast

from kasa import IotProtocol
from kasa.json import loads as json_loads
from kasa.protocols.iotprotocol import REDACTORS
from kasa.protocols.protocol import redact_data
from yarl import URL

from pykasacloud.transports import CloudTransport

_LOGGER = logging.getLogger(__name__)


class CloudProtocol(IotProtocol):
    """Cloud Protocol Class."""

    _device_id: str | None = None
    _url: URL | None = None

    def attach_device(self, device_dict: dict[str, Any]) -> None:
        """Update the appServerUrl (from a device)."""
        self._url = URL(device_dict["appServerUrl"])
        self._device_id = device_dict["deviceId"]

    async def _execute_query(self, request: str, retry_count: int) -> dict:
        debug_enabled = _LOGGER.isEnabledFor(logging.DEBUG)

        if debug_enabled:
            _LOGGER.debug(
                "%s >> %s",
                self._host,
                request,
            )
        transport: CloudTransport = cast(CloudTransport, self._transport)

        resp = await transport.send_request(
            json_loads(request), self._device_id, self._url
        )

        if debug_enabled:
            data = redact_data(resp, REDACTORS) if self._redact_data else resp
            _LOGGER.debug(
                "%s << %s",
                self._host,
                pf(data),
            )
        return resp



"""Constants for PyKasaCloud integration."""

ACCOUNT_ID = "account_id"
API_URL = "https://wap.tplinkcloud.com"
APPTYPE = "Tapo_Ios"
USERAGENT = "User-Agent: Dalvik/2.1.0 (Linux; U; Android 6.0.1; A0001 Build/M4B30X)"
TOKEN = "token"
REFRESH_TOKEN = "refresh_token"
CLIENT_ID = "client_id"
APPSERVERURL = "appServerUrl"



"""PyKasaCloud - A Python library for interacting with Kasa Cloud API."""

from importlib.metadata import version

from .exceptions import KasaCloudError
from .kasacloud import DeviceDict, KasaCloud, Token
from .protocols import CloudProtocol
from .transports import CloudTransport

__version__ = version("python-kasa")

__all__ = [
    "CloudProtocol",
    "CloudTransport",
    "DeviceDict",
    "KasaCloud",
    "KasaCloudError",
    "Token",
]



"""KasaCloud."""

from collections.abc import Callable, Coroutine
import logging
import time
from typing import Any

from aiohttp import ClientSession
from kasa import Device, DeviceType
from kasa.iot import (
    IotBulb,
    IotDevice,
    IotDimmer,
    IotLightStrip,
    IotPlug,
    IotStrip,
    IotWallSwitch,
)

from .exceptions import KasaCloudError
from .protocols import CloudProtocol
from .transports import CloudTransport, Token

type DeviceDict = dict[str, Any]

_GET_DEVICES_QUERY: dict[str, str] = {"method": "getDeviceList"}

GET_SYSINFO_QUERY: dict[str, dict[str, dict]] = {
    "system": {"get_sysinfo": {}},
}

_LOGGER = logging.getLogger(__name__)


class KasaCloud:
    """Class to instantiate and get devices."""

    _transport: CloudTransport

    @classmethod
    async def kasacloud(
        cls,
        client_session: ClientSession,
        *,
        username: str | None = None,
        password: str | None = None,
        token: dict[str, Any] | None = None,
        token_storage_file: str | None = None,
        token_update_callback: Callable[[Token], Coroutine] | None = None,
    ) -> "KasaCloud":
        """Get module kasacloud."""

        self = cls()

        ctoken: Token | None = Token(**token) if token else None

        self._transport = await CloudTransport.auth(
            client_session=client_session,
            username=username,
            password=password,
            token=ctoken,
            token_storage_file=token_storage_file,
            token_update_callback=token_update_callback,
        )

        return self

    @property
    def token(self) -> Token:
        """Return the token associated with this authentication."""
        return self._transport.token

    async def close(self) -> None:
        """Close the underlying resources."""
        await self._transport.close()

    async def get_device_list(self) -> list[DeviceDict]:
        """Get kasa device ids from cloud."""

        protocol: CloudProtocol = CloudProtocol(transport=self._transport)

        resp: dict = await protocol.query(_GET_DEVICES_QUERY)

        if "deviceList" not in resp:
            raise KasaCloudError(f"Invalid result {resp}")

        device_list: list[dict[str, Any]] = resp["deviceList"]

        return [device for device in device_list if device["status"]]

    async def get_device(self, device_dict: dict[str, Any]) -> Device:
        """Initantiate and populate the device.

        Taken from device_factory.py in python-kasa.
        """
        debug_enabled = _LOGGER.isEnabledFor(logging.DEBUG)
        if debug_enabled:
            start_time = time.perf_counter()

        def _perf_log(has_params: bool, perf_type: str) -> None:
            nonlocal start_time
            if debug_enabled:
                end_time = time.perf_counter()
                _LOGGER.debug(
                    "Device %s with connection params %s took %.2f seconds to %s",
                    device_dict["deviceId"],
                    has_params,
                    end_time - start_time,
                    perf_type,
                )
                start_time = time.perf_counter()

        device_class: type[Device] | None
        device: Device | None = None

        protocol: CloudProtocol = CloudProtocol(transport=self._transport)
        protocol.attach_device(device_dict)

        info = await protocol.query(GET_SYSINFO_QUERY)
        _perf_log(True, "get_sysinfo")

        device_class = _get_device_class_from_sys_info(info)
        device = device_class(device_dict["deviceId"], protocol=protocol)
        device.update_from_discover_info(info)
        await device.update()
        _perf_log(True, "update")
        return device


def _get_device_class_from_sys_info(sysinfo: dict[str, Any]) -> type[IotDevice]:
    """Find SmartDevice subclass for device described by passed data."""
    TYPE_TO_CLASS = {  # pylint: disable=invalid-name
        DeviceType.Bulb: IotBulb,
        DeviceType.Plug: IotPlug,
        DeviceType.Dimmer: IotDimmer,
        DeviceType.Strip: IotStrip,
        DeviceType.WallSwitch: IotWallSwitch,
        DeviceType.LightStrip: IotLightStrip,
        # Disabled until properly implemented
        # DeviceType.Camera: IotCamera,
    }
    return TYPE_TO_CLASS[IotDevice._get_device_type_from_sys_info(sysinfo)]  # pylint: disable=protected-access  # noqa: SLF001



"""Extension of python-kasa exceptions for Kasa Cloud integration."""

from enum import IntEnum

from kasa.exceptions import KasaException


class KasaCloudError(KasaException):
    """Exception raised for errors in the Kasa Cloud API interaction."""

    def __init__(self, msg: str) -> None:
        """Initialize the KasaCloudError with a message."""
        super().__init__(msg)


class MissingCredentials(KasaCloudError):
    """Exception raised when credentials are missing."""


class CloudErrorCode(IntEnum):
    """Enum for cloud error codes."""

    SUCCESS = 0
    TOKEN_EXPIRED = -20651
    MISSING_METHOD = -20103
    MISSING_PARAMETER = -20104
    MISSING_REQUEST_DATA = -20573
    DEVICE_OFFLINE = -20571
    ACCOUNT_NOT_FOUND = -20600
    INVALID_EMAIL_OR_PASSWORD = -20601
